////////////////////////////////////////////////////////////
//                             //
//               -               //
//                                           //
//                        v 0.1                           //
//                                                        //
//              Copyright (C) 2007 Sagrer                 //
//             LGPL 2.1         //
//                  .  lgpl.txt                     //
//                                                        //
//                   sagrer@yandex.ru                     //
////////////////////////////////////////////////////////////

//      , ....  :
// 1) Sagrer (sagrer@yandex.ru)

////////////////////////////////////////////////////////////////////////

unit PascalCodeModifier;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, LCLTextFileInString, CodeToolManager,
  CodeTree, CodeCache, CodeAtom;
  
type
  RPascalConst = record      //    .
    ConstStart : LongInt;
    ConstEnd : LongInt;
    Name : AnsiString;
    NameStart : LongInt;
    NameEnd : LongInt;
    Value : AnsiString;
    ValueStart : LongInt;
    ValueEnd : LongInt;
  end;

  APascalConsts = array of RPascalConst;    // .

  RPascalFunction = record       //    .
    FunctionStart : LongInt;
    FunctionEnd : LongInt;
    Name : AnsiString;
    NameStart : LongInt;
    NameEnd : LongInt;
    IsProcedure : Boolean;
    IsDeclaration : Boolean;
  end;

  APascalFunctions = array of RPascalFunction;    //  .

  TPascalCodeModifier = class (TTextFileInString)
  private
    //    .
    CodeBuf : TCodeBuffer;                //.
    ACodeTool : TCodeTool;                //.
    CodeToolsAreUnderControl : Boolean;    //    - ?
  public
    //.
    //    .
    Consts : APascalConsts;              //...
    Functions : APascalFunctions;    //...
    //.
    ErrorMessage : AnsiString;   //    ...
    ErrorPosition : LongInt;     //      .
    ErrorLine : LongInt;         //      .

    //-
    constructor Create; override;                           //  %).
    Destructor Destroy; override;                           // .
    // .
    Function Load(const FName : AnsiString) : boolean; override; //  .
    Procedure MakNewFile; override;                              //"  " %).
    Procedure GetErrorLineFromPos();                             //       .
    //--,    --
    procedure ResetConstsBlock();                                    //    .
    procedure ResetFunctionsBlock();                                 //    .
    procedure ResetBlocksInfo();                                     //    .
    procedure ResetConst(var ConstToReset : RPascalConst);           //   .
    procedure ResetFunction(var FuncToReset : RPascalFunction);      //   .
    function ParseConsts() : Boolean;                                //  ...
    function ParseFunctions() : Boolean;                             //  ...
    function ReParseFile() : Boolean;                                //  ...
    function GetPosForFirstFunction() : Integer;                     //        .
    //--    --
    function ReadCodeConstValue(const ConstName : AnsiString) : AnsiString;   //   .
    function WriteCodeConstValue(const ConstName, ConstValue : AnsiString) : Boolean;    //   .
    function ReadFunctionName(const FullFunctionText : AnsiString) : AnsiString;   //      .
    function FunctionIsExists(const FunctionName : AnsiString) : Boolean;    //  ?
    function FunctionIsExistsByText(const FullFunctionText : AnsiString) : Boolean;    //  ?    ,      .
    function GetFunctionIndex(const FunctionName : AnsiString) : Integer;    //         .
    function GetFullFunctionText(const FunctionIndex : Integer) : AnsiString;   //        .
    function GetFullFunctionText(const FunctionName : AnsiString) : AnsiString;   //      .
    function WriteFunction(const FunctionIndex : Integer; const FullFunctionText : AnsiString) : Boolean;   //   ,     - .
    function WriteFunction(const FunctionName, FullFunctionText : AnsiString) : Boolean;   //   ,     - .
  end;

implementation

//////////////////////////////////////////////
//          TPascalCodeModifier             //
//////////////////////////////////////////////

//------------------------------------//
//      -      //
//------------------------------------//

Constructor TPascalCodeModifier.Create;
//  %).
begin
  inherited;       //  ...

  //     .
  SetLength(Self.Consts, 0);
  SetLength(Self.Functions, 0);

  //   .
  Self.MakNewFile;
end;

Destructor TPascalCodeModifier.Destroy;
// .
begin
  inherited;
  Self.ResetBlocksInfo();
end;

//------------------------------------------//
//            ...             //
//------------------------------------------//

//------------------------------------------//
//             ...            //
//------------------------------------------//

Function TPascalCodeModifier.Load(const FName : AnsiString) : boolean;
//  .
begin
  Self.MakNewFile;
  Result := inherited;
end;

Procedure TPascalCodeModifier.MakNewFile;
//"  " %).
begin
  inherited;
  //   .
  Self.ErrorMessage := 'none';
  Self.ErrorPosition := 0;
  Self.ErrorLine := 0;
  //   .
  Self.ACodeTool := nil;
  Self.CodeBuf := nil;
  Self.CodeToolsAreUnderControl := false;
  //     .
  Self.ResetBlocksInfo();
end;

Procedure TPascalCodeModifier.GetErrorLineFromPos();
//       .
var
  TempCursor, CurrLine : LongInt;
begin
  if Self.ErrorPosition <= Self.Siz then begin
    //    .
    CurrLine := 1;
    TempCursor := 0;
    Self.ErrorLine := 0;
    repeat
      //             .
      TempCursor := TempCursor+1;
      if Self.FileString[TempCursor] = #10 then begin
        // .
        CurrLine := CurrLine+1;
      end;
      if Self.ErrorPosition = TempCursor then begin
        //   .
        Self.ErrorLine := CurrLine;
      end;
    until (TempCursor = Self.Siz) or (Self.ErrorLine <> 0);
  end
  else begin
    //    .
    Self.ErrorLine := 0;
  end;
end;

//---------------------------------------------//
//      --,    --         //
//---------------------------------------------//

procedure TPascalCodeModifier.ResetConstsBlock();
//   .
begin
  if Length(Self.Consts) > 0 then begin
    SetLength(Self.Consts, 0);
  end;
end;

procedure TPascalCodeModifier.ResetFunctionsBlock();
//    .
begin
  if Length(Self.Functions) > 0 then begin
    SetLength(Self.Functions,0);
  end;
end;

procedure TPascalCodeModifier.ResetBlocksInfo();
//    .
begin
  //....
  Self.ResetConstsBlock();
  //...
  Self.ResetFunctionsBlock();
end;

procedure TPascalCodeModifier.ResetConst(var ConstToReset : RPascalConst);
//   .
begin
  ConstToReset.ConstStart := 0;
  ConstToReset.ConstEnd := 0;
  ConstToReset.Name := '';
  ConstToReset.NameStart := 0;
  ConstToReset.NameEnd := 0;
  ConstToReset.Value := '';
  ConstToReset.ValueStart := 0;
  ConstToReset.ValueEnd := 0;
end;

procedure TPascalCodeModifier.ResetFunction(var FuncToReset : RPascalFunction);
//   .
begin
  FuncToReset.FunctionStart := 0;
  FuncToReset.FunctionEnd := 0;
  FuncToReset.Name := '';
  FuncToReset.NameStart := 0;
  FuncToReset.NameEnd := 0;
  FuncToReset.IsProcedure := true;
  FuncToReset.IsDeclaration := false;
end;

function TPascalCodeModifier.ParseConsts() : Boolean;
//  ...
var
  IControlCodeTools : Boolean;
  MemStream : TMemoryStream;
  CurrTreeNode, ConstSectNode, ConstDefNode : TCodeTreeNode;
  CurrConstNum : Integer;
  
begin
  //.
  Result := true;
  // -     CodeTools ?
  if Self.CodeToolsAreUnderControl = false then begin
    //  .
    Self.CodeBuf := CodeToolBoss.CreateFile('tempfile.pas');
    MemStream := TMemoryStream.Create;
    MemStream.Write(Self.FileString[1], Length(Self.FileString));
    Self.CodeBuf.LoadFromStream(MemStream);
    MemStream.Free;
    //  .
    CodeToolBoss.Explore(Self.CodeBuf, Self.ACodeTool, true, false);
    //    CodeTools.
    IControlCodeTools := true;
    Self.CodeToolsAreUnderControl := true;
  end
  else begin
    //-   .
    IControlCodeTools := false;
  end;
  
  //     .
  //     .
  Self.ACodeTool.BuildTree(false);
  CurrTreeNode := Self.ACodeTool.Tree.Root;
  //     .     ,
  // ,  ,  .
  if (CurrTreeNode.Desc = ctnProgram) or
     (CurrTreeNode.Desc = ctnPackage) or
     (CurrTreeNode.Desc = ctnLibrary) or
     (CurrTreeNode.Desc = ctnUnit) then
  begin
    //,       .     ...
    CurrTreeNode := CurrTreeNode.FirstChild;
    if CurrTreeNode <> nil then begin
      //,   .      (..  
      // ) -    .
      repeat
        if CurrTreeNode.Desc = ctnConstSection then begin
          //,  .        
          // ...
          ConstSectNode := CurrTreeNode;  //   .
          CurrTreeNode := CurrTreeNode.FirstChild;
          //   .
          repeat
            //  .
            if CurrTreeNode.Desc = ctnConstDefinition then begin
              //,  .    .
              CurrConstNum := Length(Self.Consts);
              SetLength(Self.Consts,CurrConstNum+1);
              Self.ResetConst(Self.Consts[CurrConstNum]);
              //     .
              Self.Consts[CurrConstNum].ConstStart := CurrTreeNode.StartPos;
              Self.Consts[CurrConstNum].ConstEnd := CurrTreeNode.EndPos;
              //  .  ,  ,  .
              Self.ACodeTool.CurPos.StartPos := CurrTreeNode.StartPos;
              Self.ACodeTool.CurPos.EndPos := CurrTreeNode.StartPos;
              Self.ACodeTool.CurPos.Flag := cafNone;
              Self.ACodeTool.ReadNextAtom;
              Self.Consts[CurrConstNum].Name := Self.ACodeTool.GetAtom;
              Self.Consts[CurrConstNum].NameStart := Self.ACodeTool.CurPos.StartPos;
              Self.Consts[CurrConstNum].NameEnd := Self.ACodeTool.CurPos.EndPos;
              //  ...
              ConstDefNode := CurrTreeNode;
              CurrTreeNode := CurrTreeNode.FirstChild;
              repeat
                if CurrTreeNode.Desc = ctnConstant then begin
                  //,  .
                  Self.Consts[CurrConstNum].Value := Copy(Self.FileString, CurrTreeNode.StartPos, CurrTreeNode.EndPos-CurrTreeNode.StartPos );
                  Self.Consts[CurrConstNum].ValueStart := CurrTreeNode.StartPos;
                  Self.Consts[CurrConstNum].ValueEnd := CurrTreeNode.EndPos;
                end;
                //!
                if Self.Consts[CurrConstNum].Value = '' then begin
                  CurrTreeNode := CurrTreeNode.NextBrother;
                end
                else begin
                  CurrTreeNode := nil;
                end;
              until CurrTreeNode = nil;
              //  .
              CurrTreeNode := ConstDefNode;
            end;
            //!
            CurrTreeNode := CurrTreeNode.NextBrother;
          until CurrTreeNode = nil;
          // .
          CurrTreeNode := ConstSectNode;
        end;
        //  .
        CurrTreeNode := CurrTreeNode.NextBrother;
      until CurrTreeNode = nil;
    end
    else begin
      //.. ,  ...     ,  0   .
    end;
  end
  else begin
    //.
    Self.ErrorMessage := Self.ClassName+' can work only with full Programs, Units, Packages and Libraries.';
    Self.ErrorPosition := CurrTreeNode.StartPos;
    Self.GetErrorLineFromPos();
    Result := false;
  end;
  
  //    .
  if IControlCodeTools = true then begin
    Self.CodeBuf := nil;
    Self.ACodeTool := nil;
    IControlCodeTools := false;
    Self.CodeToolsAreUnderControl := false;
  end;
end;

function TPascalCodeModifier.ParseFunctions() : Boolean;
//  ...
var
  IControlCodeTools : Boolean;
  MemStream : TMemoryStream;
  CurrTreeNode, HeaderNode  : TCodeTreeNode;
  CurrFuncNum : Integer;
begin
  //.
  Result := true;
  // -     CodeTools ?
  if Self.CodeToolsAreUnderControl = false then begin
    //  .
    Self.CodeBuf := CodeToolBoss.CreateFile('tempfile.pas');
    MemStream := TMemoryStream.Create;
    MemStream.Write(Self.FileString[1], Length(Self.FileString));
    Self.CodeBuf.LoadFromStream(MemStream);
    MemStream.Free;
    //  .
    CodeToolBoss.Explore(Self.CodeBuf, Self.ACodeTool, true, false);
    //    CodeTools.
    IControlCodeTools := true;
    Self.CodeToolsAreUnderControl := true;
  end
  else begin
    //-   .
    IControlCodeTools := false;
  end;
  
  //     .
  Self.ACodeTool.BuildTree(false);
  CurrTreeNode := Self.ACodeTool.Tree.Root;

  //     .     ,
  // ,  ,  .
  if (CurrTreeNode.Desc = ctnProgram) or
     (CurrTreeNode.Desc = ctnPackage) or
     (CurrTreeNode.Desc = ctnLibrary) or
     (CurrTreeNode.Desc = ctnUnit) then
  begin
    //,       .     ...
    CurrTreeNode := CurrTreeNode.FirstChild;
    if CurrTreeNode <> nil then begin
      //,   .      (..  
      // ) -  .
      repeat
        if CurrTreeNode.Desc = ctnProcedure then begin
          //, .    .
          CurrFuncNum := Length(Self.Functions);
          SetLength(Self.Functions,Length(Self.Functions)+1);
          Self.ResetFunction(Self.Functions[CurrFuncNum]);
          Self.Functions[CurrFuncNum].FunctionStart := CurrTreeNode.StartPos;
          Self.Functions[CurrFuncNum].FunctionEnd := CurrTreeNode.EndPos;
          // .
          CurrTreeNode := CurrTreeNode.FirstChild;
          // -    -   -   \,   - forward.
          if CurrTreeNode.NextBrother = nil then begin
            Self.Functions[CurrFuncNum].IsDeclaration := true;
          end
          else begin
            Self.Functions[CurrFuncNum].IsDeclaration := false;
          end;
          //  .
          Self.ACodeTool.CurPos.StartPos := CurrTreeNode.StartPos;
          Self.ACodeTool.CurPos.EndPos := CurrTreeNode.StartPos;
          Self.ACodeTool.CurPos.Flag := cafNone;
          Self.ACodeTool.ReadNextAtom;
          Self.Functions[CurrFuncNum].Name := Self.ACodeTool.GetAtom;
          Self.Functions[CurrFuncNum].NameStart := Self.ACodeTool.CurPos.StartPos;
          Self.Functions[CurrFuncNum].NameEnd := Self.ACodeTool.CurPos.EndPos;
          //  ?
          HeaderNode := CurrTreeNode;
          CurrTreeNode := CurrTreeNode.FirstChild;
          repeat
            if CurrTreeNode <> nil then begin
              if CurrTreeNode.Desc = ctnIdentifier then begin
                Self.Functions[CurrFuncNum].IsProcedure := false;
              end;
              //!
              if Self.Functions[CurrFuncNum].IsProcedure = true then begin
                CurrTreeNode := CurrTreeNode.NextBrother;
              end
              else begin
                CurrTreeNode := nil;
              end;
            end;
          until CurrTreeNode = nil;
          //  .
          CurrTreeNode := HeaderNode.Parent;
        end;
        //!
        CurrTreeNode := CurrTreeNode.NextBrother;
      until CurrTreeNode = nil;
    end;
  end
  else begin
    //.
    Self.ErrorMessage := Self.ClassName+' can work only with full Programs, Units, Packages and Libraries.';
    Self.ErrorPosition := CurrTreeNode.StartPos;
    Self.GetErrorLineFromPos();
    Result := false;
  end;
  
  //    .
  if IControlCodeTools = true then begin
    Self.CodeBuf := nil;
    Self.ACodeTool := nil;
    IControlCodeTools := false;
    Self.CodeToolsAreUnderControl := false;
  end;
end;

function TPascalCodeModifier.ReParseFile() : Boolean;
var
  IControlCodeTools : Boolean;
  MemStream : TMemoryStream;

begin
  //.
  Result := true;
  // -     CodeTools ?
  if Self.CodeToolsAreUnderControl = false then begin
    //  .
    Self.CodeBuf := CodeToolBoss.CreateFile('tempfile.pas');
    MemStream := TMemoryStream.Create;
    MemStream.Write(Self.FileString[1], Length(Self.FileString));
    Self.CodeBuf.LoadFromStream(MemStream);
    MemStream.Free;
    //  .
    CodeToolBoss.Explore(Self.CodeBuf, Self.ACodeTool, true, false);
    //    CodeTools.
    IControlCodeTools := true;
    Self.CodeToolsAreUnderControl := true;
  end
  else begin
    //-   .
    IControlCodeTools := false;
  end;
  
  //   ....
  Self.ResetBlocksInfo();
  
  //     .
  Result := Self.ParseConsts();
  if Result = true then Result := Self.ParseFunctions();
  
  //    .
  if IControlCodeTools = true then begin
    Self.CodeBuf := nil;
    Self.ACodeTool := nil;
    IControlCodeTools := false;
    Self.CodeToolsAreUnderControl := false;
  end;
end;

function TPascalCodeModifier.GetPosForFirstFunction() : Integer;
//        .
var
  IControlCodeTools : Boolean;
  MemStream : TMemoryStream;
  CurrTreeNode : TCodeTreeNode;
  
begin
  //.
  Result := -1;
  
  // -     CodeTools ?
  if Self.CodeToolsAreUnderControl = false then begin
    //  .
    Self.CodeBuf := CodeToolBoss.CreateFile('tempfile.pas');
    MemStream := TMemoryStream.Create;
    MemStream.Write(Self.FileString[1], Length(Self.FileString));
    Self.CodeBuf.LoadFromStream(MemStream);
    MemStream.Free;
    //  .
    CodeToolBoss.Explore(Self.CodeBuf, Self.ACodeTool, true, false);
    //    CodeTools.
    IControlCodeTools := true;
    Self.CodeToolsAreUnderControl := true;
  end
  else begin
    //-   .
    IControlCodeTools := false;
  end;

  //     .
  Self.ACodeTool.BuildTree(false);
  CurrTreeNode := Self.ACodeTool.Tree.Root;
  
  //     .
  //      Program
  if CurrTreeNode.Desc = ctnProgram then begin
    // Program -      begin -     .
    CurrTreeNode := CurrTreeNode.FirstChild;
    if CurrTreeNode <> nil then begin
      repeat
        if CurrTreeNode.Desc = ctnBeginBlock then begin
          //,  .
          Result := CurrTreeNode.StartPos;
          CurrTreeNode := nil;
        end;
        if CurrTreeNode <> nil then CurrTreeNode := CurrTreeNode.NextBrother;
      until CurrTreeNode = nil;
    end;
  end
  else begin
    // -   %)
    Self.ErrorMessage := 'TPascalCodeModifier.GetPosForFirstFunction can work only with Programs.';
  end;
  
  //    .
  if IControlCodeTools = true then begin
    Self.CodeBuf := nil;
    Self.ACodeTool := nil;
    IControlCodeTools := false;
    Self.CodeToolsAreUnderControl := false;
  end;
end;

//---------------------------------------------------//
//      --    --      //
//---------------------------------------------------//

function TPascalCodeModifier.ReadCodeConstValue(const ConstName : AnsiString) : AnsiString;
//   .
var
  I : Integer;
begin
  //.
  Result := '';
  
  // ...
  for I := 0 to Length(Self.Consts)-1 do begin
    if LowerCase(Self.Consts[I].Name) = LowerCase(ConstName) then begin
      //.
      Result := Self.Consts[I].Value;
      Break;
    end;
  end;
end;

function TPascalCodeModifier.WriteCodeConstValue(const ConstName, ConstValue : AnsiString) : Boolean;
//   .
var
  I : Integer;
begin
  //.
  Result := false;
  
  // 2   .     -    .
  //  -      ,    .
  
  // ...
  for I := 0 to Length(Self.Consts)-1 do begin
    if LowerCase(Self.Consts[I].Name) = LowerCase(ConstName) then begin
      //.   .
      Delete(Self.FileString, Self.Consts[I].ValueStart, Self.Consts[I].ValueEnd-Self.Consts[I].ValueStart);
      Insert(ConstValue, Self.FileString, Self.Consts[I].ValueStart);
      Result := true;
      Break;
    end;
  end;
  
  //     -     .
  if Result = false then begin
    Insert(#13+#10+'  '+ConstName+' = '+ConstValue+';', Self.FileString, Self.Consts[Length(Self.Consts)-1].ConstEnd);
    Result := true;
  end;
  
  //   ,     -    .
  if Result = true then begin
    Result := Self.ReParseFile();
  end;
end;

function TPascalCodeModifier.ReadFunctionName(const FullFunctionText : AnsiString) : AnsiString;
//      .
var
  TempModifier : TPascalCodeModifier;
  Parsed : Boolean;
  
begin
  //.
  Result := '';
  TempModifier := TPascalCodeModifier.Create;
  Parsed := true;
  
  //  " "     .
  TempModifier.MakNewFile;
  TempModifier.FileString := 'Program Temp;'+#13+#10+FullFunctionText+#13+#10+'begin end.';
  //.
  try
    TempModifier.ReParseFile();
  except
    //   FullFunctionText.
    Parsed := false;
    Self.ErrorMessage := 'Error, parsing FullFunctionText argument';
  end;

  //    .
  if Parsed = true then begin
    if Length(TempModifier.Functions) > 0 then begin
      Result := TempModifier.Functions[0].Name;
    end;
  end;
  
  // .
  TempModifier.Free;
end;

function TPascalCodeModifier.FunctionIsExists(const FunctionName : AnsiString) : Boolean;
//  ?
var
  I : Integer;
begin
  //.
  Result := false;
  
  // .
  for I := 0 to Length(Self.Functions)-1 do begin
    if Self.Functions[I].IsDeclaration = false then begin
      //  forward   - .. -   .
      if LowerCase(Self.Functions[I].Name) = LowerCase(FunctionName) then begin
        //.
        Result := true;
        Break;
      end;
    end;
  end;
  
end;

function TPascalCodeModifier.FunctionIsExistsByText(const FullFunctionText : AnsiString) : Boolean;
//  ?    ,      .
begin
  Result := Self.FunctionIsExists(Self.ReadFunctionName(FullFunctionText));
end;

function TPascalCodeModifier.GetFunctionIndex(const FunctionName : AnsiString) : Integer;
//         .
var
  I : Integer;
begin
  //.
  Result := -1;
  
  // .
  for I := 0 to Length(Self.Functions)-1 do begin
    if Self.Functions[I].IsDeclaration = false then begin
      //  forward   - .. -   .
      if LowerCase(Self.Functions[I].Name) = LowerCase(FunctionName) then begin
        //.
        Result := I;
        Break;
      end;
    end;
  end;
  
end;

function TPascalCodeModifier.GetFullFunctionText(const FunctionIndex : Integer) : AnsiString;
//        .
begin
  Result := Copy(Self.FileString, Self.Functions[FunctionIndex].FunctionStart, (Self.Functions[FunctionIndex].FunctionEnd-Self.Functions[FunctionIndex].FunctionStart)+1);
end;

function TPascalCodeModifier.GetFullFunctionText(const FunctionName : AnsiString) : AnsiString;
//      .
var
  FuncIndex : Integer;
begin
  //.
  Result := '';
  
  //    -    .
  FuncIndex := Self.GetFunctionIndex(FunctionName);
  //   .
  if FuncIndex >= 0 then begin
    Result := GetFullFunctionText(FuncIndex);
  end;
end;

function TPascalCodeModifier.WriteFunction(const FunctionIndex : Integer; const FullFunctionText : AnsiString) : Boolean;
//   ,     - .
begin
  //.
  Result := false;
  
  //   .
  Delete(Self.FileString, Self.Functions[FunctionIndex].FunctionStart, (Self.Functions[FunctionIndex].FunctionEnd-Self.Functions[FunctionIndex].FunctionStart)+1);
  //   .
  Insert(FullFunctionText, Self.FileString, Self.Functions[FunctionIndex].FunctionStart);
  //   .
  Result := Self.ReParseFile();
end;

function TPascalCodeModifier.WriteFunction(const FunctionName, FullFunctionText : AnsiString) : Boolean;
//   ,     - .
var
  FuncIndex, InputPos : Integer;
  TailToAdd : AnsiString;
begin
  //.
  Result := false;

  //  ?
  FuncIndex := Self.GetFunctionIndex(FunctionName);
  if FuncIndex >= 0 then begin
    //  .     .
    Result := Self.WriteFunction(FuncIndex, FullFunctionText);
  end
  else begin
    // -    .
    if Length(Self.Functions) > 0 then begin
      //     1 
      InputPos := (Self.Functions[Length(Self.Functions)-1].FunctionEnd)+1;
      TailToAdd := '';
      //    %).
      Insert(#13+#10, Self.FileString, InputPos);
      InputPos := InputPos+2;
    end
    else begin
      //     .
      InputPos := Self.GetPosForFirstFunction();
      TailToAdd := #13+#10+#13+#10;
    end;
    
    //  .
    if InputPos > 0 then begin
      Insert(#13+#10+FullFunctionText+TailToAdd, Self.FileString, InputPos);
      //   .
      Result := Self.ReParseFile();
    end;
  end;
end;

end.

