///////////////////////////////////////////////////////////
//                     GGBuildTools                      //
//         Gipat Group       //
//             Copyright (C) 2007 Gipat Group            //
//                             //
//    Gipat Group's opened EI-editor-utility license     //
//                       1.0                       //
//                                                       //
//                  www.gipatgroup.org                   //
///////////////////////////////////////////////////////////

//      , ....  :
// 1) Sagrer (sagrer@yandex.ru)

////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////
//           *.sconf       //
//                  //
//                GGBuildTools                 //
//                 1                 //
////////////////////////////////////////////////////

unit SconfFormat;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, LCLAnIniFile;

const
  ScF_Version = 1;               //    .
  //  
  ScF_ProjectType_Unknown = -1;              // .
  ScF_ProjectType_VerInfoTxtOnly = 0;        //,   GGBuildTools     *.vit     .
  ScF_ProjectType_PascalCode = 1;            //       *.pas
  ScF_ProjectType_CppCode = 2;               //       *.h
  // .
  ScF_ArchiveType_Unknown = 'Unknown';       // .
  ScF_ArchiveType_7z = '7z';                 //  *.7z.
  //    .
  //[Main]
  ScF_Def_Main_FormatVersion = ScF_Version;
  ScF_Def_Main_ProjectType = ScF_ProjectType_Unknown;
  ScF_Def_Main_ProjectVer = '0.0.0.0';
  ScF_Def_Main_UseSvn = false;
  ScF_Def_Main_ProjectRootDir = '.\';
  //[IssBuilder]
  ScF_Def_IssBuilder_FilesDir = '.\distro\';
  ScF_Def_IssBuilder_DefaultIssName = 'default.iss';
  ScF_Def_IssBuilder_OutputDir = '.\installers\';
  ScF_Def_IssBuilder_VersionsCacheDir = '.\ggbt\iss_cache\';
  ScF_Def_IssBuilder_SaveToCache = false;
  ScF_Def_IssBuilder_CreateFull = true;
  ScF_Def_IssBuilder_CreateFull_AllowUpdate = false;
  ScF_Def_IssBuilder_CreateFull_UpdVers = '0';
  ScF_Def_IssBuilder_CreateSeparateUpds = false;
  ScF_Def_IssBuilder_CreateSeparateUpds_UpdVers = '0';
  ScF_Def_IssBuilder_CreateCumulateUpdate = false;
  ScF_Def_IssBuilder_CreateCumulateUpdate_UpdVers = '0';
  //[SourcePackBuilder]
  ScF_Def_SourcePackBuilder_FilesDir = '.\source\';
  ScF_Def_SourcePackBuilder_OutputDir = '.\sources\';
  ScF_Def_SourcePackBuilder_ArchiveFormat = ScF_ArchiveType_7z;
  
type
  RSconfMainSectionInfo = record             //     [Main]
    FormatVersion : Integer;          //  
    ProjectType : Integer;            // .   (  ): Delphi, Builder, VerInfoTxtOnly, PascalCode, CppCode
    ProjectVer : AnsiString;          //     ( 3   "0.1.2"),   "Project"    3      (    ).
    UseSvn : Boolean;                 //  svn.
    ProjectRootDir : AnsiString;      // ( ..\[  *.sconf] )  (windows)    .   = ".\".
  end;
  
  RSconfIssBuilderSectionInfo = record          //     [IssBuilder]
    FilesDir : AnsiString;                      //  root          (    +  ).
    DefaultIssName : AnsiString;                //  iss-.
    OutputDir : AnsiString;                     //  root      .
    VersionsCacheDir : AnsiString;              //  root  ,     (  -  ).
    SaveToCache : Boolean;                      //       .
    CreateFull : Boolean;                       //  ,      .
    CreateFull_AllowUpdate : Boolean;           //  ,     ,      .
    CreateFull_UpdVers : AnsiString;            // All,  From:X  X - 4    -      ()  ,   -    (  flist  iss)     .
    CreateSeparateUpds : Boolean;               //      .
    CreateSeparateUpds_UpdVers : AnsiString;    // All,  From:X  X - 4    -      ()  ,   -    (  flist  iss)     .
    CreateCumulateUpdate : Boolean;             //    .
    CreateCumulateUpdate_UpdVers : AnsiString;  // All,  From:X  X - 4    -      ()  ,   -    (  flist  iss)     .

  end;
  
  RSconfSourcePackBuilderSectionInfo = record      //     [SourcePackBuilder]
    FilesDir : AnsiString;            //  root         .
    OutputDir : AnsiString;           //  root     .
    ArchiveFormat : AnsiString;       //     .
  end;

  TSconfFormat = class          //      *.sconf
  private
    // ...
    procedure ParseProjectType(const ProjectTypeStr : AnsiString);        //          .
    function GenerateProjectTypeString() : AnsiString;                    //     ,     .
  public
    //
    MainSection : RSconfMainSectionInfo;                               //    [Main]
    IssBuilderSection : RSconfIssBuilderSectionInfo;                   //    [IssBuilder]
    SourcePackBuilderSection : RSconfSourcePackBuilderSectionInfo;     //    [SourcePackBuilder]
    ErrorMessage : AnsiString;                                         //     .

    //-...
    constructor Create;
    destructor Destroy; override;
    
    // ...
    procedure Clear();                                       //      .
    function Load(const FileName : AnsiString) : boolean;    //   .
    function Save(const FileName : AnsiString) : boolean;    //     .
  end;

implementation

/////////////////////////////////////////////
//             TSconfFormat                //
/////////////////////////////////////////////

//-----------------------------------------//
//        -...      //
//-----------------------------------------//

constructor TSconfFormat.Create;
begin
  //  ...
  Self.Clear();
  ErrorMessage := 'none';

  //   ...

end;

destructor TSconfFormat.Destroy;
begin
  // 

  //  
  inherited;
end;

//------------------------------------------//
//            ...             //
//------------------------------------------//

procedure TSconfFormat.ParseProjectType(const ProjectTypeStr : AnsiString);
//          .
begin

  if LowerCase(ProjectTypeStr) = 'verinfotxtonly' then begin
    MainSection.ProjectType := ScF_ProjectType_VerInfoTxtOnly;
  end
  else if LowerCase(ProjectTypeStr) = 'pascalcode' then begin
    MainSection.ProjectType := ScF_ProjectType_PascalCode;
  end
  else if LowerCase(ProjectTypeStr) = 'cppcode' then begin
    MainSection.ProjectType := ScF_ProjectType_CppCode;
  end
  else begin
    MainSection.ProjectType := ScF_ProjectType_Unknown;
  end;

end;

function TSconfFormat.GenerateProjectTypeString() : AnsiString;
//     ,     .
begin
  //  .
  Result := '';

  //.
  if MainSection.ProjectType = ScF_ProjectType_VerInfoTxtOnly then begin
    Result := 'VerInfoTxtOnly';
  end
  else if MainSection.ProjectType = ScF_ProjectType_PascalCode then begin
    Result := 'PascalCode';
  end
  else if MainSection.ProjectType = ScF_ProjectType_CppCode then begin
    Result := 'CppCode';
  end
  else if MainSection.ProjectType = ScF_ProjectType_Unknown then begin
    Result := 'Unknown';
  end;
  
end;

//------------------------------------------//
//             ...            //
//------------------------------------------//

procedure TSconfFormat.Clear();
//      .
begin
  //  ...

  //[Main]
  MainSection.FormatVersion := ScF_Def_Main_FormatVersion;
  MainSection.ProjectType := ScF_Def_Main_ProjectType;
  MainSection.ProjectVer := ScF_Def_Main_ProjectVer;
  MainSection.UseSvn := ScF_Def_Main_UseSvn;
  MainSection.ProjectRootDir := ScF_Def_Main_ProjectRootDir;
  
  //[IssBuilder]
  IssBuilderSection.FilesDir := ScF_Def_IssBuilder_FilesDir;
  IssBuilderSection.DefaultIssName := ScF_Def_IssBuilder_DefaultIssName;
  IssBuilderSection.OutputDir := ScF_Def_IssBuilder_OutputDir;
  IssBuilderSection.VersionsCacheDir := ScF_Def_IssBuilder_VersionsCacheDir;
  IssBuilderSection.SaveToCache := ScF_Def_IssBuilder_SaveToCache;
  IssBuilderSection.CreateFull := ScF_Def_IssBuilder_CreateFull;
  IssBuilderSection.CreateFull_AllowUpdate := ScF_Def_IssBuilder_CreateFull_AllowUpdate;
  IssBuilderSection.CreateFull_UpdVers := ScF_Def_IssBuilder_CreateFull_UpdVers;
  IssBuilderSection.CreateSeparateUpds := ScF_Def_IssBuilder_CreateSeparateUpds;
  IssBuilderSection.CreateSeparateUpds_UpdVers := ScF_Def_IssBuilder_CreateSeparateUpds_UpdVers;
  IssBuilderSection.CreateCumulateUpdate := ScF_Def_IssBuilder_CreateCumulateUpdate;
  IssBuilderSection.CreateCumulateUpdate_UpdVers := ScF_Def_IssBuilder_CreateCumulateUpdate_UpdVers;
  
  //[SourcePackBuilder]
  SourcePackBuilderSection.FilesDir := ScF_Def_SourcePackBuilder_FilesDir;
  SourcePackBuilderSection.OutputDir := ScF_Def_SourcePackBuilder_OutputDir;
  SourcePackBuilderSection.ArchiveFormat := ScF_Def_SourcePackBuilder_ArchiveFormat;
  
end;

function TSconfFormat.Load(const FileName : AnsiString) : boolean;
//   .
var
  IniFile1 : TAnIniFile;
  AllOk : boolean;
  
begin
  //.
  IniFile1 := TAnIniFile.Create;
  Result := false;

  //,   ...
  AllOk := FileExists(FileName);
  if AllOk = false then begin
    //   ...
    ErrorMessage := 'File '+FileName+' is not exists.';
  end
  else begin
    //    ... ...
    AllOk := IniFile1.Load(FileName);
    if AllOk = false then begin
      //   .
      ErrorMessage := 'Error, loading file '+FileName;
    end;
  end;
  
  if AllOk = true then begin
    //, .  ...
    if IniFile1.ReadString('Main','FormatName') <> 'Sconf' then begin
      //  Sconf-.
      AllOk := false;
      ErrorMessage := 'File '+FileName+' is not Sconf file.';
    end;
  end;
  
  if AllOk = true then begin
    //  ...
    MainSection.FormatVersion := IniFile1.ReadInteger('Main','FormatVersion');
    if MainSection.FormatVersion > ScF_Version then begin
      //       .
      AllOk := false;
      ErrorMessage := 'File '+FileName+' has more new version ('+IntToStr(MainSection.FormatVersion)+') then i know ('+IntToStr(ScF_Version)+'). Aborted loading. Try update to a new version of this software.';
    end;
  end;
  
  if AllOk = true then begin
    //  ...
    //      -     .

    //[Main]
    Self.ParseProjectType(IniFile1.ReadString('Main','ProjectType'));
    MainSection.ProjectVer := IniFile1.ReadString('Main','ProjectVer');
    MainSection.UseSvn := IniFile1.ReadBool('Main','UseSvn');
    MainSection.ProjectRootDir := IniFile1.ReadString('Main','ProjectRootDir');
    
    //[IssBuilder]
    IssBuilderSection.FilesDir := IniFile1.ReadString('IssBuilder','FilesDir');
    IssBuilderSection.DefaultIssName := IniFile1.ReadString('IssBuilder','DefaultIssName');
    IssBuilderSection.OutputDir := IniFile1.ReadString('IssBuilder','OutputDir');
    IssBuilderSection.VersionsCacheDir := IniFile1.ReadString('IssBuilder','VersionsCacheDir');
    IssBuilderSection.SaveToCache := IniFile1.ReadBool('IssBuilder','SaveToCache');
    IssBuilderSection.CreateFull := IniFile1.ReadBool('IssBuilder','CreateFull');
    IssBuilderSection.CreateFull_AllowUpdate := IniFile1.ReadBool('IssBuilder','CreateFull_AllowUpdate');
    IssBuilderSection.CreateFull_UpdVers := IniFile1.ReadString('IssBuilder','CreateFull_UpdVers');
    IssBuilderSection.CreateSeparateUpds := IniFile1.ReadBool('IssBuilder','CreateSeparateUpds');
    IssBuilderSection.CreateSeparateUpds_UpdVers := IniFile1.ReadString('IssBuilder','CreateSeparateUpds_UpdVers');
    IssBuilderSection.CreateCumulateUpdate := IniFile1.ReadBool('IssBuilder','CreateCumulateUpdate');
    IssBuilderSection.CreateCumulateUpdate_UpdVers := IniFile1.ReadString('IssBuilder','CreateCumulateUpdate_UpdVers');
    
    //[SourcePackBuilder]
    SourcePackBuilderSection.FilesDir := IniFile1.ReadString('SourcePackBuilder','FilesDir');
    SourcePackBuilderSection.OutputDir := IniFile1.ReadString('SourcePackBuilder','OutputDir');
    SourcePackBuilderSection.ArchiveFormat := IniFile1.ReadString('SourcePackBuilder','ArchiveFormat');

    //,      .       .
    Result := true;
  end;
  
  // .
  IniFile1.Free;
end;

function TSconfFormat.Save(const FileName : AnsiString) : boolean;
//     .
var
  IniFile1 : TAnIniFile;
  
begin
  //  .        ...
  Result := false;   //  .

  //
  IniFile1 := TAnIniFile.Create;

  //    .
  IniFile1.MakNewFile;
  
  // ,      [Main].
  IniFile1.WriteString('Main','FormatName','Sconf');          //
  IniFile1.WriteInteger('Main','FormatVersion',ScF_Version);     // 
  
  //[Main]
  IniFile1.WriteString('Main','ProjectType',GenerateProjectTypeString());
  IniFile1.WriteString('Main','ProjectVer',MainSection.ProjectVer);
  IniFile1.WriteBool('Main','UseSvn',MainSection.UseSvn);
  IniFile1.WriteString('Main','ProjectRootDir',MainSection.ProjectRootDir);
  
  //[IssBuilder]
  IniFile1.WriteString('IssBuilder','FilesDir',IssBuilderSection.FilesDir);
  IniFile1.WriteString('IssBuilder','DefaultIssName',IssBuilderSection.DefaultIssName);
  IniFile1.WriteString('IssBuilder','OutputDir',IssBuilderSection.OutputDir);
  IniFile1.WriteString('IssBuilder','VersionsCacheDir',IssBuilderSection.VersionsCacheDir);
  IniFile1.WriteBool('IssBuilder','SaveToCache',IssBuilderSection.SaveToCache);
  IniFile1.WriteBool('IssBuilder','CreateFull',IssBuilderSection.CreateFull);
  IniFile1.WriteBool('IssBuilder','CreateFull_AllowUpdate',IssBuilderSection.CreateFull_AllowUpdate);
  IniFile1.WriteString('IssBuilder','CreateFull_UpdVers',IssBuilderSection.CreateFull_UpdVers);
  IniFile1.WriteBool('IssBuilder','CreateSeparateUpds',IssBuilderSection.CreateSeparateUpds);
  IniFile1.WriteString('IssBuilder','CreateSeparateUpds_UpdVers',IssBuilderSection.CreateSeparateUpds_UpdVers);
  IniFile1.WriteBool('IssBuilder','CreateCumulateUpdate',IssBuilderSection.CreateCumulateUpdate);
  IniFile1.WriteString('IssBuilder','CreateCumulateUpdate_UpdVers',IssBuilderSection.CreateCumulateUpdate_UpdVers);
  
  //[SourcePackBuilder]
  IniFile1.WriteString('SourcePackBuilder','FilesDir',SourcePackBuilderSection.FilesDir);
  IniFile1.WriteString('SourcePackBuilder','OutputDir',SourcePackBuilderSection.OutputDir);
  IniFile1.WriteString('SourcePackBuilder','ArchiveFormat',SourcePackBuilderSection.ArchiveFormat);

  //    .
  Result := IniFile1.Save(FileName);          //    %).
  //   -   -    ,    %).
  //        ,    ,   
  //    -          ,  
  //   ,    ,  GUI,   ....

  // 
  IniFile1.Free;
end;

end.

